import os

import numpy as np

from progressbar import progressbar
from biopandas.mol2 import PandasMol2
from sklearn.metrics.pairwise import euclidean_distances


# Set the Angstrom Euclidean distance cutoff for a protein atom to be
# considered as being from the pocket. Protein atoms that are less than CUTOFF
# Angstroms away from at least one co-crystal ligand atoms are considered to be
# pocket atoms.
CUTOFF = 10.

# Initialize the co-crystal ligand MOL2 reader.
mol2_reader = PandasMol2()


def get_ligand_coords(pdb_id):
    """Returns the co-crystal ligand's coordinates."""

    # Read in the MOL2 file and extract the coordinates from the resultant
    # dataframe.
    mol2_file = mol2_reader.read_mol2(f"../data/{pdb_id}/crystal_ligand.mol2")
    mol2_file = mol2_file.df.loc[:, ["x", "y", "z"]]

    # Return the dataframe subset as a numpy array.
    return mol2_file.values


def get_protein_coords_(line):
    """Returns the coordinates from the line as a list of floats."""

    # Keep just the last three elements of the split line, which correspond to
    # the atom's coordinates.
    coords = line.split()[-3:]

    # Return the coordinates as a list of floats.
    return [float(coord) for coord in coords]


def get_protein_coords(pdb_id):
    """Returns the protein's PDB lines and atom coordinates."""

    # Read in all the lines from the protein's PDB.
    with open(f"../data/{pdb_id}/receptor.pdb", "r") as f:
        lines = [line.strip() for line in f.readlines()]

    # Remove the lines that do not describe a protein atom.
    lines = [line for line in lines if line.split()[0] not in ("TER", "END")]

    # Remove the extraneous suffixes that exist in a few lines. They follow the
    # format 1.00%d%d%d.%d%d%d and seemingly do not describe any meaningful
    # coordinate.
    for i in range(len(lines)):
        # The length of the lines with these suffixes is always 66.
        if len(lines[i]) == 66:
            lines[i] = lines[i][:54]
        # The length of the lines without these suffixes is always 54.
        elif len(lines[i]) == 54:
            pass
        # This exception, which indicates the line is neither 54 nor 66
        # characters long, should never need to be raised when working with
        # these PDBs.
        else:
            raise Exception(f"Unexpected line: {lines[i]}")

    # Get just the coordinates from each of these lines as a list of floats.
    coords = [get_protein_coords_(line) for line in lines]

    # Return the lines and their corresponding coordinates.
    return lines, coords


def save_pocket(pdb_id):
    """Writes the pocket subset of the protein PDB into its own PDB."""

    # Get both the protein PDB lines and the coordinates of the protein and
    # ligand to find the pocket atoms from.
    protein_lines, protein_coords = get_protein_coords(pdb_id)
    ligand_coords = get_ligand_coords(pdb_id)

    # Get a Euclidean distance matrix between the protein and ligand atoms.
    dist_matrix = euclidean_distances(protein_coords, ligand_coords)
    # Convert the distance matrix to an indicator matrix that denotes when a
    # protein and ligand atom are less than CUTOFF Angstroms apart.
    dist_matrix = dist_matrix < CUTOFF

    # Sum how many ligand atoms each protein atom is less than the cutoff
    # distance from.
    line_totals = np.sum(dist_matrix, axis=1)
    # Convert these totals to an indices vector denoting which protein atoms
    # are close enough to at least one ligand atom.
    selected_indices = np.where(line_totals)[0]

    # Select just the PDB lines corresponding to these protein atoms.
    pocket_lines = [protein_lines[i] for i in selected_indices]
    # Save the subset PDB in the input folder with an indication of which
    # cutoff was used.
    with open(f"../data/{pdb_id}/receptor_pocket_{CUTOFF}.pdb", "w") as f:
        f.write("\n".join(pocket_lines))


# Save a protein pocket PDB for each DUD-E protein PDB.
for pdb_id in os.listdir("../data/"):
    save_pocket(pdb_id)

